#include "classes.cxx"
#include <header/ppaa22.amplitude.doublevirtual.h>
#include "definitions.observable.set.cxx"

void ppaa22_calculate_H2(observable_set & oset){
  static Logger logger("ppaa22_calculate_H2");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;
  
  double s = (osi_p_parton[0][1]+osi_p_parton[0][2]).m2();
  double t = (osi_p_parton[0][1]-osi_p_parton[0][4]).m2();
  double u = (osi_p_parton[0][1]-osi_p_parton[0][3]).m2();
  
//   cout << "s=" << s << ", t=" << t << ", u=" << u << endl;

  double Q_factor;
  if (osi_name_process == "uu~_aa") {
    Q_factor=pow(2.0/3,4);
  } else if (osi_name_process == "dd~_aa") {
    Q_factor=pow(1.0/3,4);
  } else {
    assert(false);
  }
  

  double Qq;
  if (osi_name_process == "uu~_aa") {
    Qq=2.0/3;
  } else if (osi_name_process == "dd~_aa") {
    Qq=-1.0/3;
  } else {
    assert(false);
  }


  osi_QT_A0 = Q_factor*ppaa22_A0(s,t,u)*pow(4*M_PI*osi_msi.alpha_e,2);
  
  osi_QT_A0 /= (4*9); // averaging factor
  osi_QT_A0 *= 0.5; // symmetry factor
  
  // H1
  osi_QT_A1 = 2*(inv2pi * osi_alpha_S)*ppaa22_H1f(s,t,u)*osi_QT_A0;
  osi_QT_H1_delta = osi_QT_A1;
  
  // H2
  osi_QT_A2 = pow(2*inv2pi * osi_alpha_S,2)*ppaa22_H2f(s,t,u,osi_N_f,Qq)*osi_QT_A0;
  osi_QT_H2_delta = osi_QT_A2;
   
  // adapt alpha_S/Pi normalisation
  osi_QT_H1_delta /= 2;
  osi_QT_H2_delta /= 4;
  
  osi_QT_H1_delta /= (inv2pi * osi_alpha_S * osi_QT_A0);
  osi_QT_H2_delta /= (pow(inv2pi * osi_alpha_S, 2) * osi_QT_A0);
  
//   cout << "H1=" << osi_QT_H1_delta << endl;
//   cout << "H2=" << osi_QT_H2_delta << endl;
       
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

