#ifndef MULTICHANNEL_SET_H
#define MULTICHANNEL_SET_H

class multichannel_set {
private:

public:
  // constant over full multichannel
  string name;
  int n_channel;
  int n_alpha_events;
  int n_alpha_steps;
  int switch_minimum_weight;
  double limit_minimum_weight;
  double reserved_minimum_weight;

  // used only inside optimization routines
  // could be private
  double a_reserved_min;

  // chanes for each phase-space point
  int channel;
  vector<double> g_channel;
  vector<double> g_IS_channel;

  // changes at optimization steps
  vector<double> alpha;
  vector<double> beta;

  // needed for optimization phase
  //  int end_optimization;

  int active_optimization;
  int end_optimization;

  int n_gen;
  int n_acc;
  int n_rej;
  int n_tec;
  int n_nan;

  vector<int> n_acc_channel;
  vector<int> n_rej_channel;
  vector<double> w_channel_step_sum;
  vector<double> w_channel_full_sum;
  //  vector<double> MC_sum_w_channel;
  vector<double> w_channel_av;

  // could be private
  int counter_minimum_weight;
  vector<int> no_channel_minimum_weight;

  int i_alpha_it;
  vector<vector<double> > alpha_it;
  vector<double> diff_w; // -> MC_diff_w_step
  int x_minimum_diff_w;


////////////////////
//  constructors  //
////////////////////
  multichannel_set();
  multichannel_set(int _n_channel);
  multichannel_set(string _name, int _n_channel, int _n_alpha_events, int _n_alpha_steps, int _switch_minimum_weight, double _limit_minimum_weight, double _reserved_minimum_weight, int _active_optimization, int _end_optimization);

  void psp_MCweight_optimization(double & integrand, double & g_tot);
  void step_MCweight_optimization(int i_acc, int i_rej);
  void result_MCweight_optimization(int i_acc, int i_rej, string & filename_MCweight);
  void output_MCweight_optimization(string & filename_MCweight);

  //  multichannel_set(multichannel_set ms);
  //  multichannel_set(int _no_prc, vector<int> _type_parton, vector<vector<int> > _pdf);
  //  multichannel_set(string _name, vector<string> _all_name, int _type_splitting, vector<int> _in_collinear, int _no_prc, vector<int> _type_parton, vector<vector<int> > _pdf, double _charge_factor, int _no_emitter, int _no_spectator, vector<int> _pair, int _type_correction, int _massive);
  //  multichannel_set(string _name, vector<string> _all_name, vector<int> _type_splitting, vector<vector<int> > _in_collinear, int _no_prc, vector<int> _type_parton, vector<vector<int> > _pdf, vector<double> _charge_factor, vector<int> _no_emitter, vector<int> _no_spectator, vector<vector<int> > _pair, vector<int> _type_correction, vector<int> _massive);
///////////////////////
//  access elements  //
///////////////////////
/*
  string name() const;
  const vector<string> all_name() const;
  int type() const;
  const vector<int> in_collinear() const;
  int no_prc() const;
  const vector<int> type_parton() const;
  //  const vector<int> pdf() const;
  const vector<vector<int> > pdf() const;
  double charge_factor() const;
  int no_emitter() const;
  int no_spectator() const;
  int no_OL_entry() const;
  const vector<int> pair() const;

  int type_correction() const;
  int massive() const;
*/
};
#endif
