#include "../include/classes.cxx"
//#include "../include/definitions.observable.set.cxx"
//  void vegasgrid_average(vector<double> & tau_alpha, vector<double> & tau_alpha_new);
//  void vegasgrid_calculation(vector<double> & tau_alpha, vector<double> & tau_beta, vector<int> & tau_cuts_channel, vector<int> & tau_counts_channel, vector<double> & sum_channel_weight, vector<double> & sum_channel_weight2);

void vegasgrid_average(vector<double> & tau_alpha, vector<double> & tau_alpha_new){
  Logger logger("vegasgrid_average");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  tau_alpha_new[0] = .75 * tau_alpha[0] + .25 * tau_alpha[1];
  for (int j = 1; j < tau_alpha.size() - 1; j++){
    tau_alpha_new[j] = .5 * tau_alpha[j] + .25 * tau_alpha[j - 1] + .25 * tau_alpha[j + 1];
  }
  tau_alpha_new[tau_alpha.size() - 1] = .75 * tau_alpha[tau_alpha.size() - 1] + .25 * tau_alpha[tau_alpha.size() - 2];
  if (tau_alpha_new.size() == 1){tau_alpha_new[0] = 1.;}

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

void vegasgrid_calculation(vector<double> & alpha, vector<double> & beta, vector<int> & cuts_channel, vector<int> & counts_channel, vector<double> & sum_channel_weight, vector<double> & sum_channel_weight2){
  Logger logger("vegasgrid_calculation");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  vector<double> averaged_channel_weight(alpha.size(), 0.);
  vector<double> averaged_channel_weight2(alpha.size(), 0.);
  for (int j = 0; j < alpha.size(); j++){
    if (counts_channel[j] + cuts_channel[j] != 0){
      averaged_channel_weight[j] = sum_channel_weight[j] / (counts_channel[j] + cuts_channel[j]);
      averaged_channel_weight2[j] = sum_channel_weight2[j] / (counts_channel[j] + cuts_channel[j]);
    }
    else {
      averaged_channel_weight[j] = 0.;
      averaged_channel_weight2[j] = 0.;
    }
  }
  vector<double> value(alpha.size(), 0.);
  for (int j = 0; j < averaged_channel_weight.size(); j++){
    //    if (channel_weight[j].size() != 0){
    value[j] = averaged_channel_weight[j]; // original version !!!
    //    double temp = (averaged_channel_weight2[j] - pow(averaged_channel_weight[j], 2) / (counts_channel[j] + cuts_channel[j])) / (counts_channel[j] + cuts_channel[j] - 1);
    //    if (temp > 0.){value[j] = sqrt(temp);} // works quite well !!!
    //    else {value[j] = 0.;}
    //    double temp = averaged_channel_weight2[j] - (averaged_channel_weight[j], 2) / channel_weight[j].size();
    //    if (temp > 0.){value[j] = pow(temp, 0.25);} // works quite well !!!
    //    else {value[j] = 0.;}
    //      value[j] = averaged_channel_weight2[j] - (averaged_channel_weight[j], 2) / channel_weight[j].size(); // works worse than with sqrt(...)
    //    }
    //    else {value[j] = 0.;}
  }
  double sum_value = accumulate(value.begin(), value.end(), 0.);
    for (int j = 0; j < averaged_channel_weight.size(); j++){
      logger << LOG_DEBUG_VERBOSE <<"value[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << value[j] << "   averaged_channel_weight[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << averaged_channel_weight[j] << "   n_acc_events = " << counts_channel[j] << "   n_rej_events = " << cuts_channel[j] << endl;
    }
  for (int j = 0; j < alpha.size(); j++){value[j] = value[j] / sum_value;}
  int MCweight_lc_min = 0;
  vector<int> MCweights_set_to_min(alpha.size());
  double MCweight_limit_min = 1.e-3 / alpha.size();
  for (int j = 0; j < value.size(); j++){
    if (value[j] < MCweight_limit_min){
      MCweight_lc_min++;
      MCweights_set_to_min[j] = 1;
      value[j] = 0.;
    }
  }
  sum_value = accumulate(value.begin(), value.end(), 0.);
  double normalization = (1. - MCweight_lc_min * MCweight_limit_min) / sum_value;
  for (int j = 0; j < value.size(); j++){
    if (MCweights_set_to_min[j] == 1){value[j] = MCweight_limit_min;}
    else {value[j] = value[j] * normalization;}
  }
  alpha = value;

  vector<double> alpha_new(alpha.size());
  vegasgrid_average(alpha, alpha_new);
  //    for (int j = 0; j < beta.size(); j++){cout << "alpha[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << showpoint << alpha[j] << "   alpha_new[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << showpoint << alpha_new[j] << endl;}
  alpha = alpha_new;

  beta.resize(alpha.size());
  beta[0] = alpha[0];
  for (int i = 1; i < beta.size(); i++){beta[i] = beta[i - 1] + alpha[i];}
  //  for (int j = 0; j < beta.size(); j++){cout << "alpha[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << showpoint << alpha[j] << "   beta[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << beta[j] << "   " << setprecision(8) << setw(15) << sqrt(h_propto_pot(double(j) / beta.size(), 0, psi_exp_pdf)) * (2 * psi_E) << "   n_accepted = " << setw(8) << channel_weight[j].size() << "   n_rejected = " << setw(8) << cuts_channel[j] << "   n_total = " << setw(8) << channel_weight[j].size() + cuts_channel[j] << endl;}

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
  }



void phasespace_set::psp_MCweight_optimization(double & integrand, double & this_psp_weight, double & this_psp_weight2, double & optimization_modifier){
  Logger logger("phasespace_set::psp_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;
  double mod_integrand = integrand * optimization_modifier;
  double mod_this_psp_weight = this_psp_weight * optimization_modifier;
  double mod_this_psp_weight2 = this_psp_weight2 * pow(optimization_modifier, 2);

  // this part could be replaced by new multichannel class !!!
  ////  MC.psp_MCweight_optimization(mod_integrand, g_tot)
  if (MC_opt_end == 0){
    counts_channel[MC_channel]++;
    double w_help = pow(abs(mod_integrand), 2) / g_tot;
    logger << LOG_DEBUG_VERBOSE << "g_tot = " << g_tot << endl;
    logger << LOG_DEBUG_VERBOSE << "mod_integrand = " << mod_integrand << endl;
    //  double w_help = pow(mod_integrand, 2) / pow(gtot, 3); // original version
    if (std::isnan(w_help)){// || w_help == 0.){
      logger << LOG_DEBUG << "std::isnan(w_help)" << endl;
      logger << LOG_DEBUG << "mod_integrand = " << mod_integrand << endl;
      logger << LOG_DEBUG << "g_tot = " << g_tot << endl;
      for (int k = 0; k < MC_n_channel; k++){logger << LOG_DEBUG << "MC_g_channel[" << k << "] = " << MC_g_channel[k] << endl;}
    }
    else {
      for (int k = 0; k < MC_n_channel; k++){
        if (std::isnan(MC_g_channel[k]) || std::isnan(w_help)) {
          logger << LOG_ERROR << "weight in channel " << k << " is nan; " << MC_g_channel[k] << ", " << w_help << endl;
        } else {
	  //          if (MC_g_channel[k] * w_help < 0.){logger << LOG_INFO << "[" << setw(2) << k << "]:   MC_g_channel = " << setw(23) << MC_g_channel[k] << ", w_help = " << setw(23) << w_help << endl;}
	  temp_sum_w_channel[k] += MC_g_channel[k] * w_help;
          logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << k << "]:   temp_sum_w_channel = " << setw(23) << temp_sum_w_channel[k] << ", MC_g_channel = " << setw(23) << MC_g_channel[k] << ", w_help = " << setw(23) << w_help << endl;
        }
      }
    }
  //    new_MCweight_opt_call(MC_g_channel, temp_sum_w_channel, abs(mod_integrand), g_tot, MC_n_channel);
  }
  ////

  // MC_tau - multichannel for partonic CMS energy tau
  MC_tau.psp_MCweight_optimization(mod_integrand, g_tot);


  //  logger << LOG_INFO << "RA_x_a = " << RA_x_a << endl;
  // RA_x_a is the dipole whose channel has been chosen for phase-space generation.

  if (type_contribution == "RA" || 
      type_contribution == "RRA"){
    // MC_x_dipole[i_a] - multichannel for partonic CMS energy in dipole kinematics
    //    if (RA_x_a > 0){MC_x_dipole[RA_x_a].psp_MCweight_optimization(mod_integrand, g_tot);}
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      MC_x_dipole[i_a].psp_MCweight_optimization(mod_integrand, g_tot);
    }
  }

  // additional multichannel stuff could be added for tau, x_dipole, etc. !!!

  if (tau_opt_end == 0){
    tau_counts_channel[tau_channel]++;
    sum_tau_channel_weight[tau_channel] += abs(mod_this_psp_weight);
    sum_tau_channel_weight2[tau_channel] += mod_this_psp_weight2;
  }
  
  if (x1x2_opt_end == 0){
    x1x2_counts_channel[x1x2_channel]++;
    sum_x1x2_channel_weight[x1x2_channel] += abs(mod_this_psp_weight);
    sum_x1x2_channel_weight2[x1x2_channel] += mod_this_psp_weight2;
  }
  // weight_IS only in RA version:
  //  if (weight_IS > 0){random_manager.optimize(mod_this_psp_weight, mod_this_psp_weight2);}
  random_manager.optimize(mod_this_psp_weight, mod_this_psp_weight2);

  if (std::isnan(mod_integrand)){logger << LOG_FATAL << "unidentified error - mod_integrand nan - run stopped!" << endl; exit(1);}
  if (abs(mod_integrand) > 1.e200){logger << LOG_FATAL << "unidentified error - mod_integrand inf - run stopped!" << endl; exit(1);}
  if (std::isnan(g_tot)){logger << LOG_FATAL << "unidentified error - gtot nan - run stopped!" << endl; exit(1);}
  //  if (abs(gtot) > 1.e200){logger << LOG_DEBUG << "unidentified error - gtot inf - run stopped!" << endl; exit(1);}
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

void phasespace_set::psp_MCweight_optimization_CA(double & integrand, double & this_psp_weight, double & this_psp_weight2){
  Logger logger("phasespace_set::psp_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (z1z2_opt_end == 0){
    for (int i_z = 1; i_z < 3; i_z++){
      //    cout << "i_z = " << i_z << endl;
      //    cout << "z1z2_channel[" << i_z << "] = " << z1z2_channel[i_z] << endl;
      //    cout << "z1z2_counts_channel[" << i_z << "].size() = " << z1z2_counts_channel[i_z].size() << endl;
      //    cout << "sum_z1z2_channel_weight[" << i_z << "].size() = " << sum_z1z2_channel_weight[i_z].size() << endl;
      //    cout << "sum_z1z2_channel_weight2[" << i_z << "].size() = " << sum_z1z2_channel_weight2[i_z].size() << endl;
      z1z2_counts_channel[i_z][z1z2_channel[i_z]]++;
      sum_z1z2_channel_weight[i_z][z1z2_channel[i_z]] += abs(this_psp_weight);
      sum_z1z2_channel_weight2[i_z][z1z2_channel[i_z]] += this_psp_weight2;
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}








//new_MCweight_opt_call(abs(osi_integrand), psi_g_tot, psi_MC_n_channel);
//void new_MCweight_opt_call(double integrand, double gtot, int n_channel){
///void phasespace_set::psp_MCweight_optimization(observable_set & oset){
/*
void phasespace_set::psp_MCweight_optimization(double & integrand){
  Logger logger("phasespace_set::psp_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;
  if (MC_opt_end != 0){return;}
  counts_channel[MC_channel]++;
  double w_help = pow(abs(integrand), 2) / g_tot;
  //  double w_help = pow(integrand, 2) / pow(gtot, 3); // original version
  if (std::isnan(w_help)){// || w_help == 0.){
    logger << LOG_DEBUG << "std::isnan(w_help)" << endl;
    logger << LOG_DEBUG << "integrand = " << integrand << endl;
    logger << LOG_DEBUG << "g_tot = " << g_tot << endl;
    for (int k = 0; k < MC_n_channel; k++){logger << LOG_DEBUG << "MC_g_channel[" << k << "] = " << MC_g_channel[k] << endl;}
  }
  else {
    for (int k = 0; k < MC_n_channel; k++){
      temp_sum_w_channel[k] += MC_g_channel[k] * w_help;
    }
  }
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}
*/

//void phasespace_set::step_MCweight_optimization(observable_set & oset){
void phasespace_set::step_MCweight_optimization(){
  Logger logger("phasespace_set::step_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;
  /*
  if (MC_opt_end != 0){return;}
  if (i_acc > n_alpha_events * n_alpha_steps){return;}
  */
  if (MC_opt_end == 0 && i_acc <= n_alpha_events * n_alpha_steps){

    for (int j = 0; j < MC_n_channel; j++){
      logger << LOG_DEBUG_VERBOSE << "[" << setw(2) << j << "]:   " << "MC_alpha = " << setw(23) << MC_alpha[j] << ", temp_sum_w_channel = " << setw(23) << temp_sum_w_channel[j] << ", sum_w_channel = " << setw(23) << sum_w_channel[j] << endl;
    }

  for (int i = 0; i < MC_n_channel; i++){
    sum_w_channel[i] += temp_sum_w_channel[i];
  }
  temp_sum_w_channel = vector<double> (MC_n_channel, 0.);
  /*
  if (i_acc % n_alpha_events != 0){return;}
  //  if (i_acc % n_alpha_events != 0 || i_acc > n_alpha_events * n_alpha_steps){return;}
  //  if (i_acc % n_alpha_events == 0 && i_acc <= n_alpha_events * n_alpha_steps){
  {
*/
  if (i_acc % n_alpha_events == 0){
    ///    ofstream out_integration;
    ///    out_integration.open(osi_filename_integration.c_str(), ofstream::out | ofstream::app);
    vector<double> w_channel_av(MC_n_channel);
    vector<double> c_w_sum(MC_n_channel);
    double c_w_norm;
    alpha_it[i_alpha_it] = MC_alpha;
    for (int j = 0; j < MC_n_channel; j++){
      w_channel_av[j] = sum_w_channel[j] / (n_alpha_events);
      /*
      // if statement should not be needed !!!
      if (counts_channel[j] + cuts_channel[j] != 0){
	//w_channel_av[j] = sum_w_channel[j] / (i_rej + i_acc);
	w_channel_av[j] = sum_w_channel[j] / (n_alpha_events);
      }
      else{w_channel_av[j] = 0.;}
      */
      c_w_sum[j] = MC_alpha[j] * sqrt(w_channel_av[j]);
    }

    for (int j = 0; j < MC_n_channel; j++){
      logger << LOG_DEBUG << "[" << setw(2) << j << "]:   " << "sum_w_channel = " << setw(23) << sum_w_channel[j] << ", w_channel_av = " << setw(23) << w_channel_av[j] << ", c_w_sum = " << setw(23) << c_w_sum[j] << ", MC_alpha = " << setw(23) << MC_alpha[j] << endl;
    }

    sum_w_channel = vector<double> (MC_n_channel, 0.);
    sort(w_channel_av.begin(), w_channel_av.end());
    logger << LOG_DEBUG << "difference = " << w_channel_av[w_channel_av.size() - 1] - w_channel_av[0] << endl;
    diff_w[i_alpha_it] = w_channel_av[w_channel_av.size() - 1] - w_channel_av[0];
    c_w_norm = accumulate(c_w_sum.begin(), c_w_sum.end(), 0.);
    logger << LOG_DEBUG << "c_w_norm = " << c_w_norm << endl;
    
    if (std::isnan(c_w_norm) || std::isinf(c_w_norm)) {
      logger << LOG_ERROR << "c_w_norm is nan/inf" << endl;

      for (int j = 0; j < MC_n_channel; j++){
	logger << LOG_ERROR << "[" << setw(2) << j << "]:   " << "sum_w_channel = " << setw(23) << sum_w_channel[j] << ", w_channel_av (sorted) = " << setw(23) << w_channel_av[j] << ", c_w_sum = " << setw(23) << c_w_sum[j] << ", MC_alpha = " << setw(23) << MC_alpha[j] << endl;
      }
      for (int j = 0; j < MC_n_channel; j++){
        logger << LOG_ERROR << j << ": " << MC_alpha[j] << ", " << sum_w_channel[j] << ", " << n_alpha_events << endl;
      }
    }
    assert(!std::isnan(c_w_norm));
    
    int no_diff_w_max=0;
    double diff_w_max = 1.E300;
    for (int j = 0; j < i_alpha_it; j++){if (diff_w[j] < diff_w_max){diff_w_max = diff_w[j]; no_diff_w_max = j;}}
    MCweight_lc_min = 0;
    //    int MCweight_lc_min = 0;
    vector<int> MCweights_set_to_min;
    //    double a_reserved_min = 0.;
    a_reserved_min = 0.;
    
    if ((i_alpha_it == 0) || (diff_w[i_alpha_it] <= diff_w_max)){
      for (int j = 0; j < MC_n_channel; j++){
	MC_alpha[j] = c_w_sum[j] / c_w_norm;
      }
      // weight_min == 0: There is no minimum weight.
      // if weight_limit_min > 0 smaller weights are set to 0 (saves computation time).
      // set to zero only at the end of WO !!! (experiment)
      
      if (weight_min == 0){
	for (int j = 0; j < MC_alpha.size(); j++){
	  if (MC_alpha[j] < weight_limit_min){
	    MC_alpha[j] = 0.;
	    MCweight_lc_min++;
	  }
	}
      }
      else {
	for (int j = 0; j < MC_alpha.size(); j++){
	  if (MC_alpha[j] <= weight_limit_min){
	    MC_alpha[j] = 0.;
	    MCweights_set_to_min.push_back(j);
	    MCweight_lc_min++;
	  }
	}
      }
      
      if (MCweights_set_to_min.size() != 0 && weight_min != 0){a_reserved_min = 0.001;}
      double a_norm;
      a_norm = accumulate(MC_alpha.begin(), MC_alpha.end(), 0.) / (1. - a_reserved_min);
      for (int j = 0; j < MC_alpha.size(); j++){if (MC_alpha[j] != 0.){MC_alpha[j] = MC_alpha[j] / a_norm;}}
      for (int j = 0; j < MCweights_set_to_min.size(); j++){MC_alpha[MCweights_set_to_min[j]] = a_reserved_min / MCweight_lc_min;}
      MC_beta[0] = MC_alpha[0];
      for (int i = 1; i < MC_beta.size(); i++){MC_beta[i] = MC_beta[i - 1] + MC_alpha[i];}
    }
    else {
      MC_alpha = alpha_it[no_diff_w_max + 1];
      logger << LOG_DEBUG << "optimization step doesn't seem to improve the set of weights!" << endl;
    }
    
    ///    out_integration << "************************************************ weight optimization step " << setw(2) << i_alpha_it + 1 << " ***********************************************" << endl;
    ///    if (MCweight_lc_min != 0){
    ///      out_integration << "   " << MCweight_lc_min << " alpha's are set to " << a_reserved_min / MCweight_lc_min << "." << endl;
    ///    }
        ///    out_integration << "   D_min[alpha(i),alpha(j)](step " << i_alpha_it + 1 << ") = " << setw(20) << setprecision(15) << diff_w[i_alpha_it] << endl;
    i_alpha_it++;
    
    ///    if (!(i_acc == n_alpha_steps * n_alpha_events)){
    ///      if (MCweight_lc_min != 0){
    ///	out_integration << "   Remaining " << MCweight_lc_min << " alpha's are set to " << a_reserved_min / MCweight_lc_min << "." << endl;
    ///      }
    ///    out_integration << "****************************************************************************************************************************" << endl;
    ///    }
    ///    out_integration.close();

    for (int j = 0; j < MC_alpha.size(); j++){
      if (MC_alpha[j] == 0.){MC_g_channel[j] = 0.;}
    }
  }
  }

  logger << LOG_DEBUG << "before MC_tau.step_MCweight_optimization" << endl;
  // MC_tau - multichannel for partonic CMS energy tau
  MC_tau.step_MCweight_optimization(i_acc, i_rej);
  logger << LOG_DEBUG << "after MC_tau.step_MCweight_optimization" << endl;

  if (type_contribution == "RA" || 
      type_contribution == "RRA"){
    // MC_x_dipole[i_a] - multichannel for partonic CMS energy in dipole kinematics
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      logger << LOG_DEBUG << "MC_x_dipole[" << i_a << "].n_channel = " << MC_x_dipole[i_a].n_channel << endl;
      //    for (int i_a = 1; i_a < n_dipoles; i_a++){
      MC_x_dipole[i_a].step_MCweight_optimization(i_acc, i_rej);
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

///void phasespace_set::result_MCweight_optimization(observable_set & oset, int & size_proc_generic, vector<int> & size_proceeding){
void phasespace_set::result_MCweight_optimization(){
  Logger logger("phasespace_set::result_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (MC_opt_end != 0){return;}
  if (i_acc != n_alpha_events * n_alpha_steps){return;}
  ///    ofstream out_integration;
  ///    out_integration.open(osi_filename_integration.c_str(), ofstream::out | ofstream::app);
  vector<double> vd_ch(MC_n_channel);
  // changed from Dmin = 0, for (int j = 0... !!!
  Dmin = 1;
  ///    out_integration << "******************************************************** weight optimization result *******************************************************" << endl;
  for (int j = 0; j < diff_w.size(); j++){
    ///      out_integration << "diff_w[" << j << "] = " << setprecision(16) << setw(25) << diff_w[j] << "   " << alpha_it[j][0] << "   " << alpha_it[j][alpha_it[j].size() - 1] << endl;
    if (diff_w[j] < diff_w[Dmin]){Dmin = j;}
  }
  MC_alpha = alpha_it[Dmin];
  

  // setting weights to zero only at end of WO
  int MCweight_lc_min = 0;
  vector<int> MCweights_set_to_min;
  double a_reserved_min = 0.;
  if (weight_min == 0){
    for (int j = 0; j < MC_alpha.size(); j++){
      if (MC_alpha[j] < weight_limit_min){
	MC_alpha[j] = 0.; 
	MCweight_lc_min++;
      }
    }
  }
  else {
    for (int j = 0; j < MC_alpha.size(); j++){
      if (MC_alpha[j] <= weight_limit_min){
	MC_alpha[j] = 0.;
	MCweights_set_to_min.push_back(j);
	MCweight_lc_min++;
      }
    }
  }
  if (MCweights_set_to_min.size() != 0 && weight_min != 0){a_reserved_min = 0.001;}
  double a_norm;
  a_norm = accumulate(MC_alpha.begin(), MC_alpha.end(), 0.) / (1. - a_reserved_min);
  for (int j = 0; j < MC_alpha.size(); j++){if (MC_alpha[j] != 0.){MC_alpha[j] = MC_alpha[j] / a_norm;}}
  for (int j = 0; j < MCweights_set_to_min.size(); j++){MC_alpha[MCweights_set_to_min[j]] = a_reserved_min / MCweight_lc_min;}
  
  MC_beta = vd_ch;
  for (int j = 0; j < MC_alpha.size(); j++){
    for (int k = 0; k < j + 1; k++){
      MC_beta[j] += MC_alpha[k];
    }
  }
  /*
// warmup phase deactivated !!!
  if (weight_opt == 2){
    i_acc = 0;
    i_rej = 0;
    i_gen = 0;
    i_nan = 0;
    i_tec = 0;
    ///    osi_full_sum_weight = 0.;
    ///    osi_full_sum_weight2 = 0.;
  }
   */
 
  ///    out_integration << "******************************************************** weight optimization result *******************************************************" << endl;
  ///    out_integration << "   min(D_min[alpha(i),alpha(j)]) = " << setw(20) << setprecision(15) << diff_w[Dmin] << " at optimization step "<< setw(2) << Dmin + 1 << endl;
  ///    out_integration << "********************************************************* end of weight optimization ******************************************************" << endl;
  MC_opt_end = 1;
  ///    out_integration.close();

  /*
  //  ofstream out_MCweights(filename_MCweight.c_str());
  ofstream out_MCweights;
  //  out_integration.open(filename_MCweight.c_str(), ofstream::out | ofstream::app);
  out_MCweights.open(filename_MCweight.c_str(), ofstream::out | ofstream::trunc);
  out_MCweights << "# " << "MC" << endl;
  for (int j = 0; j < MC_alpha.size(); j++){
    out_MCweights << setprecision(16) << MC_alpha[j] << endl;
  }
  out_MCweights.close();
  MC_opt_end = 1;
  ///  size_proceeding[2] = 0;
  ///  size_proc_generic = accumulate(size_proceeding.begin(), size_proceeding.end(), 0);


  // MC_tau - multichannel for partonic CMS energy tau
  MC_tau.result_MCweight_optimization(i_acc, i_rej, filename_MCweight);

  if (type_contribution == "RA" || type_contribution == "RRA"){
    // MC_x_dipole[i_a] - multichannel for partonic CMS energy in dipole kinematics
    for (int i_a = 1; i_a < n_dipoles; i_a++){
      MC_x_dipole[i_a].result_MCweight_optimization(i_acc, i_rej, filename_MCweight);
    }
  }
  */
  logger << LOG_DEBUG_VERBOSE << "before weight_output" << endl;

  weight_output_MCweight_optimization();

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

void phasespace_set::weight_output_MCweight_optimization(){
  Logger logger("phasespace_set::result_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  //  ofstream out_MCweights(filename_MCweight.c_str());
  ofstream out_MCweights;
  //  out_integration.open(filename_MCweight.c_str(), ofstream::out | ofstream::app);
  out_MCweights.open(filename_MCweight.c_str(), ofstream::out | ofstream::trunc);
  out_MCweights << "# " << "MC" << endl;
  for (int j = 0; j < MC_alpha.size(); j++){
    out_MCweights << setprecision(16) << MC_alpha[j] << endl;
  }
  out_MCweights.close();
  MC_opt_end = 1;
  ///  size_proceeding[2] = 0;
  ///  size_proc_generic = accumulate(size_proceeding.begin(), size_proceeding.end(), 0);


  // MC_tau - multichannel for partonic CMS energy tau
  //  MC_tau.result_MCweight_optimization(i_acc, i_rej, filename_MCweight);
  if (switch_MC_tau != -1){
    MC_tau.output_MCweight_optimization(filename_MCweight);
  }

  if ((switch_MC_x_dipole != -1) &&(type_contribution == "RA" || 
				    type_contribution == "RRA")){
    // MC_x_dipole[i_a] - multichannel for partonic CMS energy in dipole kinematics
    //    for (int i_a = 1; i_a < n_dipoles; i_a++){
    for (int i_a = 1; i_a < MC_x_dipole.size(); i_a++){
      MC_x_dipole[i_a].output_MCweight_optimization(filename_MCweight);
      //      MC_x_dipole[i_a].result_MCweight_optimization(i_acc, i_rej, filename_MCweight);
    }
  }
 
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


///void phasespace_set::step_tau_weight_optimization(observable_set & oset, int & size_proc_generic, vector<int> & size_proceeding){
void phasespace_set::step_tau_weight_optimization(){
  Logger logger("phasespace_set::step_tau_weight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (tau_opt_end != 0){return;}
  if (i_acc % n_tau_events != 0 || i_acc == 0){return;}
  vegasgrid_calculation(tau_alpha, tau_beta, tau_cuts_channel, tau_counts_channel, sum_tau_channel_weight, sum_tau_channel_weight2);
  /*
  for (int i = 0; i < tau_alpha.size(); i++){
    cout << "tau_alpha[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << tau_alpha[j] << "   averaged_channel_weight[" << setw(3) << j << "] = " << setprecision(8) << setw(15) << averaged_channel_weight[j] << "   n_acc_events = " << counts_channel[j] << "   n_rej_events = " << cuts_channel[j] << endl;
  }
  */

  if (i_acc == n_tau_events * n_tau_steps){
    output_weight_vegas(filename_tauweight, tau_alpha);
    tau_opt_end = 1;
    ///    size_proceeding[3] = 0;
    ///    size_proc_generic = accumulate(size_proceeding.begin(), size_proceeding.end(), 0);
    //    logger << LOG_DEBUG << "size_proc_generic = " << size_proc_generic << endl;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

///void phasespace_set::step_x1x2_weight_optimization(observable_set & oset, int & size_proc_generic, vector<int> & size_proceeding){
void phasespace_set::step_x1x2_weight_optimization(){
  Logger logger("phasespace_set::step_x1x2_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (x1x2_opt_end != 0){return;}
  if (i_acc % n_x1x2_events != 0 || i_acc == 0){return;}
  vegasgrid_calculation(x1x2_alpha, x1x2_beta, x1x2_cuts_channel, x1x2_counts_channel, sum_x1x2_channel_weight, sum_x1x2_channel_weight2);
  if (i_acc == n_x1x2_events * n_x1x2_steps){
    output_weight_vegas(filename_x1x2weight, x1x2_alpha);
    x1x2_opt_end = 1;
    ///    size_proceeding[3] = 0;
    ///    size_proc_generic = accumulate(size_proceeding.begin(), size_proceeding.end(), 0);
    //    logger << LOG_DEBUG << "size_proc_generic = " << size_proc_generic << endl;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

///void phasespace_set::step_z1z2_weight_optimization(observable_set & oset, int & size_proc_generic, vector<int> & size_proceeding){
void phasespace_set::step_z1z2_weight_optimization(){
  Logger logger("phasespace_set::step_z1z2_MCweight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (z1z2_opt_end != 0){return;}
  if (i_acc % n_z1z2_events != 0 || i_acc == 0){return;}
  //    logger << LOG_DEBUG << "z1z2 : vegasgrid_calculation @ " << i_acc << " events" << endl;
  for (int i_z = 1; i_z < 3; i_z++){
    logger << LOG_DEBUG << "z(" << i_z << ")  : vegasgrid_calculation @ " << i_acc << " events" << endl;
    vegasgrid_calculation(z1z2_alpha[i_z], z1z2_beta[i_z], z1z2_cuts_channel[i_z], z1z2_counts_channel[i_z], sum_z1z2_channel_weight[i_z], sum_z1z2_channel_weight2[i_z]);
   }
  if (i_acc == n_z1z2_events * n_z1z2_steps){
    for (int i_z = 1; i_z < 3; i_z++){
      output_weight_vegas(filename_z1z2weight[i_z], z1z2_alpha[i_z]);
    }
    z1z2_opt_end = 1;
    // size_proceeding ???
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

