# RadISH user manual



![radish_logo_small](./share/radish_logo_small.png) (**Rad**iation off **I**nitial **S**tate **H**adrons) is a fortran90 code that resums the large logarithms for transverse observables in hadronic collisions up to N$^3$LL accuracy and performs the matching to fixed order up to N$^3$LO.

The latest version can be obtained from

```bash
git clone ...
```

## Installation requirements

RadISH depends on the following packages:

- LHAPDF (available at https://lhapdf.hepforge.org)
- HOPPET (available at https://github.com/gavinsalam/hoppet)
- CHAPLIN (available at https://chaplin.hepforge.org)

You may or may not have these packages installed on your machine. If you decide to compile RadISH using CMake (see below) and you do not have these packages, they will be automatically downloaded and installed for you. If you prefer to compile RadISH using a more standard Makefile you will need to have these packages installed and in your `PATH`.

##Compilation

To install RadISH you will need a fortran compiler. The code has been tested with several version of GNU fortran compiler (version 4.8 and higher), which is widespread and available on various operating systems. Other compilers such as ifort are currently not supported.

To compile RadISH, you have two choices.

### CMake

The simplest option is to build the program using CMake (3.0.0 or newer), as all the installation requirements will be automatically checked and if something is missing will be installed. Once you have downloaded RadISH create a build directory within the main folder

```bash
mkdir build
```

Move to the build directory and type

```bash
cmake ..
```

and

```bash
make -j
```

Which will automatically install any of the required packages (if not found) and then will build RadISH. By typing 

```bash
ccmake ..
```

before compiling a menu will appear, where you can choose among additional options available at this stage. 

### Makefile

There is also a more standard installation using a simple `Makefile`. This requires you to have already installed LHAPDF, HOPPET and CHAPLIN which have to be in your `PATH`. You can adjust their path (as well as the compilers) at the beginning of the Makefile. To build RadISH type

```bash
make configure && make -j
```

##Basic usage

The simplest way to run RadISH is through its python interface with the command

```bash
./python/radish.py -i INFILE.yml -n [number of cores]
```

where `INFILE` is a input card, written in yaml format, which contains the input parameter for each run. A template `yaml`card can be found in the `input` folder. The information on the number of cores is particularly useful when one wants to compute scale variations, which are then computed simultaneously on `n` cores (since the maximum number of scale variations is 9, you should keep `n`$\leq$ 9). 

An example of input card can be found in the folder `tutorial` which contains instruction to produce results with RadISH.

Here we briefly summarise all the options.

- ###Global run options

  These options are rather general and allow one to select the process and the observable to resum, as well as the output folder where the results are collected.  It also allows one to modify the default values for the event generation.

```yaml
run:
   process:   		can be either H, DY, Wp, Wm
   observable:		can be either
   								ptB [boson pT]
                  ptj [leading jet pT]
                  phi [phistar angle in neutral DY]
                  JV_ptB (JV_ptJ) boson pT (leading jet pT) with a veto on the
                  leading jet pT (boson pT)
   output_folder: the folder where the results will be stored
   run_name:      the name of the result files
   onshell:       set to true to compute H or NC DY on-shell production (no decay)

generation:
   seed:          the random seed
   born_events:   the number of Born events
   shower_events: the number of shower events
```

- ###Collider options

  Collider settings such as beams, energy and PDFs.

```yaml
collider:
   beams:         can be either "pp" or "ppbar"
   Ecm:           the centre-of-mass-energy of the collision [GeV]
   pdf:
      set:        the pdf set
      member:     the pdf member to use
```

- ### Calculation options

  Resummation and scale settings

```yaml
resummation:
   order:                 can be NLL, NNLL, N3LL
   modified_logs_exp:     the exponent of the modified logs which turn off the
                          resummation in the hard region of the spectrum. The default
                          value is 4. Choose between 1 and 5.
   differential:          set to true to compute the differential spectrum. If
                          false the code prints the cumulative cross section
   prime:                 set to true to compute the results with primed accuracy.
                          If true, the results will be saved appending 
                          a 'p' string to the resummed order.
   jetveto:               parameters relevant for the jetveto and the joint
   			                  resummation
      jet_radius:         the jet radius
      jet_veto:           the value of the jetveto.

scales:
   scale_settings:        can be either dynamical or fixed
   variations:            if none, only one scale is computed. If correlated, muF
                          and muR are varied simultaneously by a factor of two. If
                          all, muF and muR are varied independently (3 or 7 scale
                          variations). On top of these variations, the resummation
                          scale is varied by a factor of res_scale_vary around the central
                          scale, for a total of 5 or 9 variations
   xmuF:                  the central value of the factorisation scale. If scale
   muF:                   settings is dynamical, specify a ratio with xmuF. If
                          scale settings is fixed, specify a value.
   xmuR:     		          the central value of the renormalisation scale. If scale
   muR:                   settings is dynamical, specify a ratio with xmuF. If
                          scale settings is fixed, specify a value.
   xQ:                    the central value of the resummation scale. If scale
   Q:                     settings is dynamical, specify a ratio with xmuF. If
                          scale settings is fixed, specify a value.
   res_scale_vary:        the resummation scale variation factor (default 2)
   ren_scale_vary:        the renormalisation scale variation factor (default 2)
   fact_scale_vary:       the factorisation scale variation factor (default 2)
```

- ###Matching settings

  The details of the matching, as well as the location of the fixed order files to construct the matched predictions.
  
  Examples of fixed order files are provided in the folder `tutorial`. If you want to use your own fixed order files for matching they need to be formatted properly. The format of the fixed order files is described below.

```yaml

#

# Please insert fixed order files in the following order:

# correlated: [muF,muR], [2*muF,2*muR], [0.5*muF,0.5*muR]
# all: [muF,muR], [2*muF,2*muR], [0.5*muF,0.5*muR],
# [2*muF,muR], [muF,2*muR], [0.5*muF,muR], [muF,0.5*muR]

#-------------------------------------------------------------------------------

matching:
   order:     		        the matching order. Can be LO, NLO, NNLO
   scheme:                the matching scheme. Possible schemes are multiplicative [multa]
                        	or additive [R].
   fixed_order_file:      The fixed order file(s) used to match the predictions.
                          If empty, no matching is performed.                         
                          If scales::variations = correlated (all),
                          fixed_order_file must be a list of fixed order files of
                          length 3 (7).
                          The fixed order files *must* be inserted in the array in
                          the correct order.
                          if correlated, [muF,muR], [2*muF,2*muR],
                                         [0.5*muF,0.5*muR]
                          if all, [muF,muR], [2*muF,2*muR], [0.5*muF,0.5*muR],
                                  [2*muF,muR], [muF,2*muR], [0.5*muF,muR],
                                  [muF,0.5*muR]
   normalise:             if true, the results are normalised to the fixed-order
                          cross section.
```



- User cuts

  Basic cuts on rapidity and invariant mass. See below for additional cuts on Born-level variables.

```yaml

cuts:
   nocuts:           set to true to turn off the analysis cuts (default: true)
   boson_mass:
     min:            minimum value of the virtual boson mass
     max:            max value of the virtual boson mass
   boson_rapidity:
     min:            minimum value of the virtual boson rapidity
     max:            maximum value of the virtual boson rapidity
```



- Extra options

  If running without fixed order, these settings allow one to modify the binning. Irregular binning is currently not supported when RadISH is run without fixed order.
  
```yaml
  binning:
     obsmin:      minimum value
     obsmax:      maximum value
     nbins:       number of bins
     binscale:    can be lin [default], log (natural logarithmic) or log10.
```

## Advanced usage

### Running RadISH from command line

RadISH can also be run directly from command line, with the command

```bash
./bin/radishf90 [Options]
```

The possible options (which mirror those used in the yaml card) are displayed running the command

```bash
./bin/radishf90 -help
```

An example of the command lines generated from using the python interface can be seen using the command

```bash
./python/radish.py -d
```

Note that running RadISH from command line allows one to have access to a few additional options, such as the Higgs decay mode (by default, the Higgs does not decay), which can be useful in conjunction with additional cuts at the analysis level.

### User cuts

Users can implement their own cuts on Born variables by editing the routine `outside_cuts` in `src/src_mcfm/Generator`. Some examples are already provided both for NC and CC DY and for $H\rightarrow \gamma \gamma$ decay, and some useful routines are already implemented.

###Master Integrals

RadISH uses a set of pre-tabulated master integrals to compute the expansion of the resummation. They are stored in the folder `MI_grids`. For single-differential resummation they are available for values of the exponent $p$ of the modified logarithms between 1 and 5. For double-differential resummation they are only available for $p=4$ and values of the jet radius equal to $R=0.4$ or $R=0.8$. If you are interested in other values, please contact the developers of the code.

## Fixed order files and their format

We now briefly describe the format of the fixed order files that RadISH needs as an input when matching to fixed order. Examples of fixed order files are included in the tarball and can be found in the `tutorial` folder. 

We recall that the resummation and matching up to NNLO of generic colour singlets for all the observables available in RadISH is possible through the MATRIX+RadISH interface available at https://matrix.hepforge.org/. We encourage you to use that interface, which is described in detail in the Appendix of the MATRIX+RadISH paper:

[Accurate single- and double-differential resummation of colour-singlet processes with MATRIX+RadISH: $W^+W^-$ production at the LHC](https://inspirehep.net/literature/1791519), S. Kallweit, E. Re, L. Rottoli, M. Wisemann: arXiv:2004.07720

If you want to match the resummed predictions provided by RadISH with your own fixed order predictions the fixed order file must be formatted properly.

A basic input files should contain at least the following information in its header:

```
xsct_lo = 637738.876239                   # the LO cross section
delta_xsct_nlo = 116083.45845699997       # the NLO correction  (if available)
delta_xsct_nnlo = 17879.231331999996      # the NNLO correction (if available)
delta_xsct_nnnlo = 0.0                    # the N3LO correction (if available)
```

The following information also *must* be provided, either by including them in the header of the file (if using RadISH from command line) or in the yaml card (note that RadISH does not allow these parameters to be defined twice, so if you use RadISH with the yaml car do not put them in the fixed order file):

```yaml
# radish.f90 flag                            # yaml card counterpart
proc = DY                                    run::process
collider = pp                                collider::beams
sqrts = 13000	                               collider::Ecm
mmin = 66.0                                  cuts::boson_mass::min
mmax = 116.0                                 cuts::boson_mass::max
etamin = -2.5                                cuts::boson_rapidity::min
etamax = 2.5                                 cuts::boson_rapidity::max
pdf_name = NNPDF31_nnlo_as_0118.LHgrid       collider::pdf::set
xmur = 0.5                                   scales::xmuR
xmuf = 0.5                                   scales::xmuF
xQ = 0.5                                     scales::xQ
```
In the example above, the scales are dynamical; if the scale are fixed, you should pass the central values of the scales. If the scales are dynamical, you are responsible to pass either the dynamic flag to the RadISH executable or to set it to true in the yaml card.

You should also check the consistency of the information provided, and in particular you are responsible to set the cuts correspondingly. You might find useful to pass the flag `-dryrun` to`radish.f90`, which will allow you to compare the fixed order Born cross section you get from the Born-level event generator of RadISH to the fixed-order Born cross section.

The fixed order file then contains four columns. The first column is simply the set of the lower bin edges of the observable being resummed (e.g. $p_\perp^i$ in the case of the transverse momentum). The second, the third and the fourth column are defined as minus the integral of the differential distribution from each value of $p_t^i$ to infinity, namely
$$
\bar \Sigma(p_\perp^i) = - \int_{p_\perp^i}^\infty dp_\perp \frac{d \sigma}{d p_\perp}
$$
In particular, the second column is the anti-cumulant of the LO differential distribution, the third column is the anti-cumulant of the pure NLO correction, and the last column is the anti-cumulant of the pure NNLO correction. The files should read like this:

```
1.0 -1456174.404297349 573743.0933971701 245434.2916340814
2.0 -1039274.489864349 133987.50994516996 154317.7870120814
3.0 -828337.016355349 -1688.2345088300208 94579.36602398138
4.0 -693617.689510349 -58248.06203733002 59235.46972058137
...
```

Note that the lower edge of the first bin must be a number larger than zero.

Since fixed order predictions produced by a fixed order code are often obtained at the differential level, we provide a simple python script (you can find it in the folder `tutorial/fixed_order_example`) to convert differential distributions into a format compliant to RadISH. The script can be run with the command

```bash
./python/construct_fixed_order.py -i <differential_distributions> -b <LO_xsec> -f <NLO_xsec> -s <NNLO_xsec> -t <N3LO_xsec> -o <fixed_order_file_name>.fxd
```

where `<differential_distributions>` is a plain text file whose columns are

```
# lower_bin_edge  LO_distribution  NLO_distribution NNLO_distribution 
```

Where at least the LO distribution ought to be non-zero. Note that the script expects the NLO and NNLO distributions, not the pure NLO and NNLO corrections.

Since the anti-cumulant is defined as an integral from $p_\perp^i$ to infinity, the file which contains the differential distributions should also contain information on the overflow bin. You should therefore add to the file an extra line with this information:

```
#overflow: 0.13738  0.08527 0.01267
```

where the three numbers correspond to the LO, NLO and NNLO overflows.

## Plotting

Once the results have been obtained, they can be conveniently plotted with the command

```bash
./python/radish_plotter.py -i card.yaml
```

The following plotting options are possible:

```yaml
plot:
   order:                     Array of the desired curves to be plotted. Accuracy:
                              NLL,NLLp,NNLL,NNLLp,N3LL,LO,NLO,NNLO,LOexp,NLOexp
                              NNLOexp, and matched results, es. NLL+NLO. Make sure
                              that the file with the desired accuracy exists. For
                              fixed order and expanded results, the code will look
                              for any matched file with the desired accuracy
                              to fin the fixed order counterpart.
   data_file:                 Optional data file to be compared with data
   show_error_band:           If results with multiple scales are present, set to
                              true to plot the uncertainty band
   show_ratio:                Set to true to display the ratio with respect to a
                              selected curve
   normalise_ratio_to:        can be either central (each order is normalised to
                              its central value) or one of the variations above.
   title:                     Title of the plot
   differential:
      show:                   If true, differential distributions will be plotted
      x_label:                x label of differential plots
      y_label:                y label of cumulative plots
   cumulative:
      show:                   If true, cumulative distributions will be plotted
      x_label:                x label of differential plots
      y_label:                y label of differential plots
   options:
      set_logx:               set to true to use log scale
      set_logy:               set to true to use log scale
      xmin:                   min value on the x axis
      xmax:                   max value on the x axis
      ymin:                   min value on the y axis
      ymax:                   max value on the y axis
      ymin_ratio:             min value on the y axis of the ratio plots
      ymax_ratio:             max value on the y axis of the ratio plots
      drawstyle:              options are default (line), steps-mid (steps)
   output:                    the name of the output file
```

In addition to these options, the following keys must be included to retrieve the fixed order files (alternatively, you can simply append the plotting options to the same runcard you used to produce the results.)

```yaml
run:
   output_folder: the folder where the results will be stored
   run_name:      the root name of the result files

 scales:
   variations:    none, all or correlated
```

##Citation Policy

If you use RadISH for a scientific publication you *must* cite

[Higgs Transverse-Momentum Resummation in Direct Space](https://inspirehep.net/literature/1444881), P. Monni, E. Re, P. Torrielli: Phys.Rev.Lett. 116 (2016) no.24, 242001

[Momentum-space resummation for transverse observables and the Higgs $p_\perp$ at N$^3$LL+NNLO](https://inspirehep.net/literature/1601293), W. Bizon, P. Monni, E. Re, L. Rottoli, P. Torrielli: JHEP 1802 (2018) 108

If you use this program also for the joint resummation of pT and pTj, you should also cite:   

[Higgs Transverse Momentum with a jet veto: a double differential resummation](https://inspirehep.net/literature/1753579), P. Monni, L. Rottoli, P. Torrielli, arXiv:1909.04704



You are also encouraged to cite the original references to the programs on which RadISH relies. 

PDFs are handled through thanks to the LHAPDF interface

[LHAPDF6: parton density access in the LHC precision era](https://inspirehep.net/literature/1335438), Eur.Phys.J. C75 (2015) 3, 132, arXiv:1412.7420

The evolution of PDFs and coefficient functions and their convolution is performed by HOPPET

[A Higher Order Perturbative Parton Evolution Toolkit (HOPPET)](https://inspirehep.net/literature/784080), Comput.Phys.Commun. 180 (2009), arXiv:0804.3755

HPLs are evaluated numerically by CHAPLIN and by hplog

[CHAPLIN - Complex Harmonic Polylogarithms in Fortran](https://inspirehep.net/literature/916030),  Comput.Phys.Commun. 185 (2014), arXiv:1106.5739

[Numerical Evaluation of Harmonic Polylogarithms](https://inspirehep.net/literature/560030), Comput.Phys.Commun. 141(2001), [0107173] hep-ph/0107173

Finally, the Born matrix elements for H and DY production are obtained from MCFM

[Vector boson pair production at the LHC](https://inspirehep.net/literature/897881), JHEP 1107, 018 (2011), arXiv:1105.0020



Finally, you are also encouraged to cite the original references for the ingredients necessary to achieve N$^3$LL accuracy for $p_\perp$ and $\phi^*$, and NNLL accuracy for $p_{\perp}^J$.

