      subroutine mcfm_vegas(myinit,myitmx,myncall,mybin,xinteg,xerr)
      implicit none
      include 'types.f'
************************************************************************
*                                                                      *
*  This routine should perform the sweeps of vegasnr                   *
*                                                                      *
*    Input parameters:                                                 *
*       myinit  :  the vegasnr routine entry point                     *
*       myitmx  :  the number of vegasnr sweeps                        *
*      myncall  :  the number of iterations per sweep                  *
*          bin  :  whether or not the results should be histogrammed   *
*                                                                      *
*    Returned variables:                                               *
*       xinteg  :  value of integration                                *
*         xerr  :  integration error                                   *
*                                                                      *
*    This version is modified to work with photon processes.           *
*    At NLO one can have:                                              *
*                                                                      *
*    virt -> Born + virtual corrections + integrated QCD counterterms  *
*    real -> Real + subtraction terms                                  *  
*                [+ QED photon-quark subtractions if frag=.true.]      *
*    frag -> Fragmentation contributions, incl. integrated QED dipoles *
*                                                                      *
*    tota -> virt + real [frag=.false.]                                *
*    tota -> virt + real + frag [frag=.true.]                          *
*                                                                      *
************************************************************************

      include 'vegas_common.f'

      integer myitmx,myinit,i,j,k,mynproc,nprocabove
      integer(kind=8) myncall,myncall_save
      integer:: ierr
      logical:: mybin,bin
      real(dp):: sig,sd,chi,sigr,sdr,sigdk,sddk,chidk,
     & sigfrag,sdfrag,chifrag,sigWdk,sdWdk,chiWdk,
     & xreal,xreal2,xinteg,xerr,adjust,myscale,myfacscale,
     & mymb,sumsig,sumsigr,sumsigf,sumsd,sumsdr,sumsdf,
     & sigips(4),sdips(4),xcallwt
      integer mykpart
      character*3 getstr,psgen
      character*15 kpartstring
      common/mykpart/mykpart
      common/bin/bin
      common/xreal/xreal,xreal2
      real(dp):: lowint,virtint,realint,fragint,scetint
      real(dp):: region(2*mxdim),lord_bypart(-1:1,-1:1)
      logical:: first,myreadin
      common/bypart/lord_bypart
      external lowint,virtint,realint,fragint,scetint
      data first/.true./
      save first,sigips,sdips
 
c--- Initialize all integration results to zero, so that the
c--- total of virt and real may be combined at the end for ktota
      sig=0d0
      sd=0d0

c--- Put the vegasnr parameters in the common block
      itmx=myitmx
      ncall=myncall
      bin=mybin
      

c---  Basic lowest-order integration
c      if (kpart==klord) then
         call boundregion(ndim,region)
         call vegasnr(region,ndim,lowint,myinit,myncall,myitmx,
     &        nprn,sig,sd,chi)
c      endif

      
c--- calculate integration variables to be returned
      xinteg=sig
      xerr=sqrt(sd**2)
      

      return
      end
      
      
c#######################################################################
      subroutine boundregion(idim,region)
      implicit none
      include 'types.f'
c--- Initializes integration region [0,1] for each variable
c--- in the idim-dimensional integration range
      
      include 'mxdim.f'
      integer:: i,idim
      real(dp):: region(2*mxdim)
      
      do i=1,idim
      region(i)=0d0
      region(i+idim)=1d0
      enddo
      
      return
      end
      
      
